/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=A
  CPP_LANG=C+OPENMP
  CPP_PROCESSES_PER_NODE=jupiter 1
*/

/* RECORD
Juan Pedro López Navarro
student of the University of Murcia, Spain
January 23, 2016
time 2186 msec
speed-up 16.44
the previous record was in saturno, with 24 cores, and speed-up 19.08
and this is in jupiter, with 12 cores
*/ 

/*
Additionally, some time is improved by statically calculations of variables that do not vary
instead of making them by each loop (position calculation in memory of a row or i * t) (instead of t * t do multiplication just t times)
Static arrays are created for better memory management regarding the creation of dynamic memory or pointers.
Also they exchange operations to calculate "dc" reversing the sign, avoiding entering the "else" area more times that in the if.
Finally many threads as the size of the problem are created because after several tests, it is more efficient to create them, than the
management is to allocate more than one iteration of the loop for a thread
*/

/*
Adicionalmente, se mejora algo el tiempo haciendo cálculos de manera estática de las variables que no varían
en lugar de hacerlos por cada iteración del bucle (cálculo de la posición en memoria de una fila o i*t) (pasa de t*t multiplicaciones
a solo t veces) 
Los arrays se crean estáticos para una mejor gestión de la memoria respecto a la creación de memoria dinámica o con
punteros. 
También se intercambian las operaciones para calcular "dc" invirtiendo el signo, evitando que entre en la zona del else más veces
que en la del if.
Finalmente se crean tantos threads como el tamaño del problema ya que tras distintas pruebas es más eficiente la creación de éstos que la
gestión que se hace de asignar más de una iteración del bucle for a un hilo
*/

#include <stdlib.h>
#include <omp.h>

int unadistancia(int t,int *a,int numero)
{
  int i,j,cont=0;       //i y j servirán para índices de bucles, cont servirá como contador para ver cuantas apariciones tiene el número
  int entero;           //servirá como temporal para no rehacer cálculos
  int tam = t*t;        //también servirá al mismo propósito que "entero"
  int d=0,dist,dc;      //d llevará el valor máximo encontrado, dist representará la distancia de filas temporalmente y dc con las columnas
  
  int filas[tam];       //array estático para almacenar las posicones de las filas donde haya una aparición de numero
  int columnas[tam];    //array estático para almacenar las posicones de las columnas donde haya una aparición de numero
  
  for(i=0;i<t;i++)
  {
    entero = i*t;       //cálculo temporal de ésta operación para realizarla menos veces
    for(j=0;j<t;j++)
    {
      if(a[entero+j]==numero) //se ha encontrado una aparición del número
      {
        filas[cont]=i;        //guardamos en qué fila apareció
        columnas[cont]=j;     //guardamos en qué columna apareció
        cont++;               //incrementamos el número de apariciones
      }
    }
  }
  for(i=0;i<cont;i++)
  {
    for(j=i+1;j<cont;j++)
    {  
      dist=filas[j]-filas[i];     //calculamos la distancia de una fila a otra
      dc=columnas[j]-columnas[i]; //calculamos de una columna a otra (puede ser negativo)
      if(dc>0)                    //si es positivo sumamos
        dist+=dc;
      else                        //si es negativo lo restamos (-(-x) = +x)
        dist-=dc;
      if(dist>d)                  //si el valor nuevo es mejor actualizamos la distancia máxima
        d=dist;
    }
  }
  return d;   
}
    
void sec(int t,int *a,int n,int *numeros,int *distancias)
{
  int i;
  omp_set_num_threads(n);  //Al crear más hilos se consigue una mejora, ya que parece que el sistema trabaja mejor asignando
                            //un trabajo pequeño a muchos hilos que un trabajo grande dinámico a pocos hilos
  #pragma omp parallel for 
  for(i=0;i<n;i++)          //Con éste bucle vamos repartiendo la tarea de sacar las distancias Manhattan para cada número dado
    distancias[i]=unadistancia(t,a,numeros[i]); 
}

